import React, { Component } from 'react';
import classNames from 'classnames';
import { connect } from 'dva-no-router';
import './TreeNode.less';
import Checkbox from '../Checkbox';
import arrowUp from '../../assets/image/arrow_up.svg';
import arrowDown from '../../assets/image/arrow_down.svg';
import arrowRightPC from '../../assets/image/caret-right.svg';
import arrowDownPC from '../../assets/image/caret-down.svg';
import { loop, loadNodeChild, getChangedNodes, nodeInteractionState } from '../../utils/tree';
import { rollNodeToWhere, loadUserPromise } from '../../utils/treePC';

let loadChildDataPromise = null;
const { addPromises, getPromises } = loadUserPromise();

class TreeNode extends Component {
  constructor(props) {
    super(props);

    this.state = {
      renderFlag: false,
      loadChild: false, // 是否正在加载子节点
      hasLoadedChild: false, // 是否已经加载过子节点
      expanded: false, // 是否自动
    };

    this.renderContent = this.renderContent.bind(this);
    this.handleFoldNode = this.handleFoldNode.bind(this);
    this.handleCheck = this.handleCheck.bind(this);
    this.handleParentUpdate = this.handleParentUpdate.bind(this);
  }

  componentDidMount() {
    this.initialData(this.props);
    setTimeout(() => {
      const isExpand = this.isExpandNode();
      if (isExpand) {
        this.expandNode();
      }
    }, 0);
  }

  componentWillReceiveProps(nextProps) {
    this.initialData(nextProps);
  }

  // 是否默认展开当前节点
  isExpandNode() {
    const { treeNode, treeProps } = this.props;
    const { expandRootNode } = treeProps;
    const { _parentIdList } = treeNode;

    return expandRootNode && Array.isArray(_parentIdList) && !_parentIdList.length;
  }

  // 展开当前节点
  expandNode() {
    const { hasLoadedChild } = this.state;
    const { treeProps } = this.props;
    const { loadData } = treeProps;
    this.setState({
      expanded: true,
    });
    if (!hasLoadedChild && typeof loadData === 'function') {
      loadChildDataPromise = this.loadChildData();
      addPromises(loadChildDataPromise);
    }
  }

  initialData(props) {
    const { treeNode, treeProps, updateCheckedNodes, updateIntermediateNodeKeys } = props;
    const { _id, _checked } = treeNode;
    const {
      dynamicCheckedId,
      dynamicUnCheckedId,
      expandedNodeKeys,
      intermediateNodeKeys,
      checkedNodes,
      staticNodeKey,
      dynamicNodeKey,
      staticNodeTypeLabel,
      dynamicNodeTypeLabel,
    } = treeProps;
    const expanded = expandedNodeKeys.some(item => item === _id);

    if (updateIntermediateNodeKeys && !_checked) {
      const intermediated = nodeInteractionState({
        type: 'intermediate',
        node: treeNode,
        comparateProp: treeNode._type === staticNodeTypeLabel ? staticNodeKey : dynamicNodeKey,
        intermediateNodeKeys,
      });
      if (intermediated) {
        treeNode._intermediate = true;
      }
    }

    if (updateCheckedNodes && !_checked) {
      const checked = nodeInteractionState({
        type: 'checked',
        node: treeNode,
        nodeType: treeNode._type,
        comparateProp: treeNode._type === staticNodeTypeLabel ? staticNodeKey : dynamicNodeKey,
        attachedComparateProp: treeNode._type === dynamicNodeTypeLabel ? staticNodeKey : '',
        checkedNodes,
      });
      if (checked) {
        this.handleCheck();
      }
    }

    if (_id === dynamicCheckedId) {
      if (!_checked) {
        this.handleCheck();
      }
      const promises = getPromises();
      Promise.all(promises).finally(() => {
        this.scrollTree();
      });
    }

    if (_id === dynamicUnCheckedId) {
      if (_checked) {
        this.handleCheck();
      }
    }

    if (expanded) {
      this.expandNode();
    }
  }

  // 将树滚动至对应位置
  scrollTree() {
    const ele = this.$treeNode;
    const wrapEle = document.getElementsByClassName('tree-wrap')[0];

    rollNodeToWhere({ ele, wrapEle });
  }

  // 节点的展开/折叠
  handleFoldNode(e) {
    const { expanded, hasLoadedChild } = this.state;
    const { treeProps } = this.props;
    const { loadData } = treeProps;

    if (e && typeof e.stopPropagation === 'function') {
      e.stopPropagation();
    }

    if (typeof loadData !== 'function') {
      return this.setState({
        expanded: !expanded,
      });
    }

    if (expanded) {
      return this.setState({
        expanded: false,
      });
    }

    if (hasLoadedChild) {
      return this.setState({
        expanded: true,
      });
    }

    this.loadChildData();
  }

  // 加载子节点数据
  loadChildData() {
    const { loadChild } = this.state;
    const { treeNode, treeProps } = this.props;
    const { _checked, _isLoaded } = treeNode;
    const {
      loadData,
      loadDataFlag,
      nodeFoldDetailProps,
      checkedNodes,
      disabledNodeKeys,
      loopStyle,
    } = treeProps;
    const params = {};

    if (_isLoaded || loadChild) {
      return false;
    }

    nodeFoldDetailProps.forEach(key => {
      params[key] = treeNode[key] || '';
    });

    if (treeNode[loadDataFlag]) {
      return this.setState({
        expanded: true,
        hasLoadedChild: true,
      });
    }

    if (typeof loadData !== 'function') {
      return false;
    }

    this.setState({
      loadChild: true,
    });

    return loadData(params)
      .then(childList => {
        loadNodeChild({
          node: treeNode,
          childList,
          treeProps,
          checkedNodes,
          disabledNodeKeys,
        });

        treeNode._isLoaded = true;

        this.setState({
          hasLoadedChild: true,
        });

        // 当 loopStyle 为 entire 时，展开非叶子节点，需要重新计算 changedNodes
        if (_checked && loopStyle === 'entire') {
          setTimeout(() => {
            this.handleNodesChange();
          }, 0);
        }
      })
      .catch(err => {
        throw new Error(`loadData err: ${err}`);
      })
      .finally(() => {
        this.setState({
          expanded: true,
          loadChild: false,
        });
      });
  }

  handleNodesChange() {
    const { treeNode, _tree, treeProps, onNodesChange } = this.props;
    const { onChange, checkedNodeProps, dynamicNodeTypeLabel, filterFunc, loopStyle } = treeProps;

    const { checkedNodes, intermediateNodes } = getChangedNodes({
      checkedNodeProps,
      dynamicNodeTypeLabel,
      _tree,
      checkedNodes: treeProps.checkedNodes,
      filterFunc,
      loopStyle,
    });

    const callbackData = {
      node: treeNode,
      checkedNodes,
      intermediateNodes,
    };

    if (typeof onNodesChange === 'function') {
      onNodesChange(callbackData);
    }
    if (typeof onChange === 'function') {
      onChange(callbackData);
    }
  }

  // 选中/取消选中 节点
  handleCheck() {
    /* eslint-disable */
    const { renderFlag } = this.state;
    const { treeNode, notifyParentUpdate, treeProps } = this.props;
    const { disabledNodeKeys, deptNodeCheckable } = treeProps;
    const { _checked, _children, _id, _type } = treeNode;
    const state = !_checked;

    if (disabledNodeKeys.some(item => item === _id)) {
      return false;
    }

    if (_type === 'dept' && !deptNodeCheckable) {
      return false;
    }

    // 更新自身节点状态
    treeNode._checked = state;
    treeNode._intermediate = false;

    // 更新后代节点状态
    loop({ list: _children }, item => {
      item._intermediate = false;
      item._checked = state;
      return {
        result: item._isLeaf ? null : item._children,
      };
    });

    // 通知父节点更新状态
    if (typeof notifyParentUpdate === 'function') {
      notifyParentUpdate();
    }

    this.setState({
      renderFlag: !renderFlag,
    });

    setTimeout(() => {
      this.handleNodesChange();
    }, 0);
    /* eslint-enable */
  }

  // notifyParentUpdate 事件处理
  handleParentUpdate() {
    if (loadChildDataPromise && typeof loadChildDataPromise.then === 'function') {
      loadChildDataPromise.finally(() => {
        this.updateSelfState();
      });
    } else {
      this.updateSelfState();
    }
  }

  // 更新自身状态（由后代节点通知）
  updateSelfState() {
    const { renderFlag } = this.state;
    const { treeNode, notifyParentUpdate } = this.props;
    const { _children } = treeNode;
    let isAllChildChecked = true;
    let isAllChildUnChecked = true;

    // 确定自身状态
    for (let i = 0, item; (item = _children[i]); i += 1) {
      if (item._intermediate) {
        isAllChildChecked = false;
        isAllChildUnChecked = false;
        break;
      }
      if (item._checked) {
        isAllChildUnChecked = false;
      } else {
        isAllChildChecked = false;
      }
    }
    treeNode._checked = isAllChildChecked;
    treeNode._intermediate = !isAllChildChecked && !isAllChildUnChecked;

    // 将更新请求冒泡到父节点
    if (typeof notifyParentUpdate === 'function') {
      notifyParentUpdate();
    }

    this.setState({
      renderFlag: !renderFlag,
    });
  }

  renderCheckbox() {
    const { treeNode, treeProps } = this.props;
    const { platform, showCheckbox, deptNodeCheckable, themeImg } = treeProps;
    const { loadChild } = this.state;
    const { _disabled, _type } = treeNode;

    if (!showCheckbox || (!deptNodeCheckable && _type === 'dept')) {
      return null;
    }

    return (
      <Checkbox
        checked={treeNode._checked}
        intermediate={treeNode._intermediate}
        disabled={_disabled}
        loading={loadChild}
        styleType={platform}
        themeImg={themeImg}
      />
    );
  }

  renderFoldIcon() {
    const { treeNode, treeProps } = this.props;
    const { expanded } = this.state;
    const { platform, isTag } = treeProps;
    if (isTag) {
      return '';
    }
    return (
      !treeNode._isLeaf && (
        <div
          className={
            platform === 'mobile'
              ? classNames('tree-node-icon-wrap-mobile')
              : classNames('tree-node-icon-wrap-pc')
          }
          onClick={this.handleFoldNode}
        >
          {platform === 'mobile' && (
            <img
              className={classNames('tree-node-icon-mobile')}
              src={expanded ? arrowUp : arrowDown}
              alt="arrow"
            />
          )}
          {platform === 'pc' && (
            <img
              className={classNames('tree-node-icon-pc')}
              src={expanded ? arrowDownPC : arrowRightPC}
              alt="arrow"
            />
          )}
        </div>
      )
    );
  }

  renderMobile() {
    const { treeNode, treeProps } = this.props;
    const cls = classNames(['tree-node-wrap', 'tree-node-wrap-mobile']);
    if (treeProps.isTag) {
      treeNode.istag = treeProps.isTag;
    }
    return (
      <div className={cls} onClick={this.handleCheck}>
        {this.renderCheckbox()}
        <div style={{ width: '100%' }}>
          {treeProps.renderNodeText && typeof treeProps.renderNodeText === 'function' ? (
            treeProps.renderNodeText(treeNode)
          ) : (
            <span className={classNames('tree-node-label')} style={{ maxWidth: '150px' }}>
              {treeNode._name}
            </span>
          )}
        </div>
        {this.renderFoldIcon()}
      </div>
    );
  }

  renderPC() {
    const { treeNode, treeProps } = this.props;
    const cls = classNames(['tree-node-wrap', 'tree-node-wrap-pc']);
    if (treeProps.isTag) {
      treeNode.istag = treeProps.isTag;
    }
    return (
      <div
        className={cls}
        onClick={this.handleCheck}
        ref={ref => {
          this.$treeNode = ref;
        }}
      >
        {this.renderFoldIcon()}
        {this.renderCheckbox()}
        <div style={{ width: '100%' }}>
          {treeProps.renderNodeText && typeof treeProps.renderNodeText === 'function' ? (
            treeProps.renderNodeText(treeNode)
          ) : (
            <span className={classNames('tree-node-label')} style={{ maxWidth: '190px' }}>
              {treeNode._name}
              {treeProps.isTag ? `(${treeNode.count || 0})` : ''}
            </span>
          )}
        </div>
      </div>
    );
  }

  // 渲染子节点
  renderChildren() {
    const { expanded } = this.state;
    const {
      treeNode,
      _tree,
      updateCheckedNodes,
      updateIntermediateNodeKeys,
      treeProps = {},
      resetHighlightId,
      onNodesChange,
    } = this.props;
    const { _isLeaf, _children } = treeNode;
    const className = _isLeaf || !expanded ? 'tree-node-children-hidden' : 'tree-node-children';

    return (
      <div className={classNames(className)}>
        {!_isLeaf &&
          _children.map(item => {
            const { _id, role = '' } = item;
            return (
              <TreeNode
                key={_id + role}
                className="tree-node-component"
                treeNode={item}
                treeProps={treeProps}
                _tree={_tree}
                updateCheckedNodes={updateCheckedNodes}
                updateIntermediateNodeKeys={updateIntermediateNodeKeys}
                notifyParentUpdate={this.handleParentUpdate}
                resetHighlightId={resetHighlightId}
                onNodesChange={onNodesChange}
              />
            );
          })}
      </div>
    );
  }

  renderContent() {
    const { className, treeProps } = this.props;
    const { platform } = treeProps;

    const treeNodeCls = classNames(['nbugs-component-tree-node', className]);

    return (
      <div className={treeNodeCls}>
        {platform === 'mobile' ? this.renderMobile() : this.renderPC()}
        {/* 子节点 */}
        {this.renderChildren()}
      </div>
    );
  }

  render() {
    return this.renderContent();
  }
}

export default connect(({ global, ...rest }) => ({
  _tree: rest.tree._tree,
  updateCheckedNodes: rest.tree.updateCheckedNodes,
  updateIntermediateNodeKeys: rest.tree.updateIntermediateNodeKeys,
}))(TreeNode);
