// 允许重新赋值函数参数
/* eslint no-param-reassign: ["error", { "props": false }] */

import _cloneDeep from 'lodash/cloneDeep';

/**
 *
 * 确定某个节点的选择状态
 * @param {*} { type, node }
 */
export function nodeInteractionState({
  type = '',
  node = null,
  nodeType,
  comparateProp = '',
  attachedComparateProp = '',
  checkedNodes,
  intermediateNodeKeys,
  disabledNodeKeys,
} = {}) {
  if (!type || !node || !comparateProp) {
    throw new Error('nodeInteractionState err');
  }

  const id = node[comparateProp] || '';

  if (type === 'checked') {
    if (typeof checkedNodes[0] === 'object') {
      return checkedNodes.some(item => {
        if (item.type !== nodeType) {
          return false;
        }
        if (attachedComparateProp && item[attachedComparateProp] && node[attachedComparateProp]) {
          return (
            `${node[attachedComparateProp]}${id}` ===
            `${item[attachedComparateProp]}${item[comparateProp]}`
          );
        }
        return item[comparateProp] === id;
      });
    } else if (typeof checkedNodes[0] === 'string') {
      return checkedNodes.some(item => item === id);
    }
  } else if (type === 'intermediate') {
    return intermediateNodeKeys.some(item => item === id);
  } else if (type === 'disabled') {
    return disabledNodeKeys.some(item => item === id);
  }
}

/**
 *
 * 确定节点的选择状态
 */
function _deptNodeSelectState(_tree) {
  loop(
    { list: _tree, parentItem: null },
    item => {
      const { _children, _isLeaf } = item;

      return {
        result: _isLeaf ? null : _children,
        parentItem: item,
      };
    },
    curNode => {
      if (!curNode) {
        return false;
      }
      const { _checked, _children } = curNode;
      let _allUnChecked = true;
      let _allChecked = true;
      let intermediateFlag = false;

      if (_checked) {
        return false;
      }

      _children.forEach(childItem => {
        if (childItem._intermediate) {
          intermediateFlag = true;
        }
        if (childItem._checked) {
          _allUnChecked = false;
        } else {
          _allChecked = false;
        }
      });

      curNode._intermediate = intermediateFlag || !(_allUnChecked || _allChecked);
      curNode._checked = _allChecked;
    }
  );
}

/**
 *
 * 深度遍历
 * @param {*} list
 * @param {*} parentItem
 * @param {*} func
 */
export function loop({ list, parentItem }, func, parentNodeFunc) {
  for (let i = 0; i < list.length; i += 1) {
    const item = list[i];
    const next = typeof func === 'function' ? func(item, parentItem) : { result: null };
    if (next.result !== null) {
      loop({ list: next.result, parentItem: next.parentItem }, func, parentNodeFunc);
    }
  }
  if (typeof parentNodeFunc === 'function') {
    parentNodeFunc(parentItem);
  }
}

/**
 * 初始化树结构数据，增加组件辅助属性
 * @param {*} sourceList
 * @param {*} props
 */
export function initData({ sourceList, props }) {
  const {
    staticNodeLabel,
    child,
    staticNodeKey,
    showCheckbox,
    checkedNodes,
    loadDataFlag,
    staticNodeTypeLabel,
    intermediateNodeKeys,
    disabledNodeKeys,
    loadData,
    dynamicNodeTypeLabel,
    dynamicNodeKey,
    dynamicNodeLabel,
  } = props;
  const tempList = _cloneDeep(sourceList);

  loop({ list: tempList, parentItem: null }, (item, parentItem) => {
    if (item[dynamicNodeLabel]) {
      item._type = dynamicNodeTypeLabel;
      item._id = item[dynamicNodeKey];
      item._children = [];
      item._name = item[dynamicNodeLabel];
      item._isLeaf = true;
    } else {
      item._type = staticNodeTypeLabel;
      item._id = item[staticNodeKey];
      item._children = _cloneDeep(Array.isArray(item[child]) ? item[child] : []);
      item._name = item[staticNodeLabel];
      item._isLeaf =
        typeof loadData === 'function' && !item[loadDataFlag] ? false : !item._children.length;
    }
    // item._type = staticNodeTypeLabel;
    // item._id = item[staticNodeKey];
    // item._name = item[staticNodeLabel];
    item._parentIdList =
      parentItem && Array.isArray(parentItem._parentIdList)
        ? parentItem._parentIdList.concat(parentItem._id || '')
        : [];

    item._isLoaded = typeof loadData === 'function' && !item[loadDataFlag] ? false : null;
    item._allChecked = true;
    item._allUnChecked = true;
    item._disabled =
      nodeInteractionState({
        type: 'disabled',
        node: item,
        comparateProp: staticNodeKey,
        disabledNodeKeys,
      }) || (parentItem ? parentItem._disabled : false);

    if (showCheckbox) {
      if (parentItem && parentItem._checked) {
        item._checked = true;
        item._intermediate = false;
      } else {
        item._checked = nodeInteractionState({
          type: 'checked',
          node: item,
          nodeType: item._type,
          comparateProp: staticNodeKey,
          checkedNodes,
        });
        item._intermediate = nodeInteractionState({
          type: 'intermediate',
          node: item,
          comparateProp: staticNodeKey,
          intermediateNodeKeys,
        });
      }
    } else {
      item._checked = null;
      item._intermediate = null;
    }

    // 删除原有子节点数据，避免数据冗余
    delete item[child];

    return {
      result: item._isLeaf ? null : item._children,
      parentItem: item,
    };
  });

  // 不获取子节点时，需要对非叶子节点的选择状态进行计算
  if (loadData === null) {
    _deptNodeSelectState(tempList);
  }
  return tempList;
}

/**
 * 补充checkedNodes
 * @param {*} checkedNodesResult
 */
function extendCheckedNodes({ checkedNodesResult = [], checkedNodes, dynamicNodeTypeLabel }) {
  const result = [];

  if (!Array.isArray(checkedNodes) || !checkedNodes.length || typeof checkedNodes[0] === 'string') {
    return result;
  }

  const users = checkedNodes.filter(item => item.type === dynamicNodeTypeLabel);

  if (!users.length) {
    return result;
  }

  users.forEach(item => {
    const { userId = '', _hasloaded } = item;
    if (!checkedNodesResult.some(node => node.userId === userId) && !_hasloaded) {
      result.push(item);
    }
  });

  return result;
}

/**
 * 完成节点选择
 */
export function getChangedNodes({
  checkedNodeProps,
  // dynamicNodeTypeLabel,
  // checkedNodes,
  _tree,
  filterFunc = null,
  loopStyle = 'default',
  uncheckNode,
  callbackUncheck,
} = {}) {
  const checkedNodesResult = [];
  const intermediateNodes = [];

  if (!checkedNodeProps.length) {
    throw new Error('checkedNodeProps 长度不可为 0');
  }

  // 取消选中项
  if (uncheckNode && Object.keys(uncheckNode).length) {
    handleTest(_tree, uncheckNode);

    function handleTest(data, detail, isok) {
      data.map(item => {
        if (isok) {
          item._allChecked = !1;
          item._checked = !1;
          item._allUnChecked = !1;
          if (item._children && item._children.length) {
            handleTest(item._children, detail, isok);
          }
          return true;
        }
        if (detail.deptName === item.deptName && item._type === detail.type) {
          if (item._type === 'user' && item.userId === detail.userId) {
            item._intermediate = !1;
            item._checked = !1;
            return true;
          } else if (item._type === 'dept' && detail.deptId === item.deptId) {
            item._allChecked = !1;
            item._checked = !1;
            // eslint-disable-next-line no-unused-expressions
            item._children && item._children.length && handleTest(item._children, detail, !0);
          }
        }
        if (item._children && item._children.length) {
          handleTest(item._children, detail);
        }
        return true;
      });
    }

    if (typeof callbackUncheck === 'function') {
      callbackUncheck();
    }
  }

  loop({ list: _tree }, item => {
    const { _type } = item;
    const tempItem = {
      type: _type,
    };

    checkedNodeProps.forEach(key => {
      tempItem[key] = item[key] === undefined ? '' : item[key];
    });

    const { _checked, _intermediate, _isLeaf, _children } = item;
    if (_checked) {
      if (typeof filterFunc === 'function') {
        if (filterFunc(_cloneDeep(tempItem))) {
          checkedNodesResult.push(tempItem);
        }
      } else {
        checkedNodesResult.push(tempItem);
      }
      if (loopStyle === 'default') {
        return { result: null };
      }
    }

    if (_intermediate) {
      intermediateNodes.push(tempItem);
    }

    return {
      result: _isLeaf ? null : _children,
    };
  });

  return {
    // checkedNodes: checkedNodesResult.concat(
    //   extendCheckedNodes({
    //     intermediateNodes,
    //     checkedNodesResult,
    //     checkedNodes,
    //     dynamicNodeTypeLabel,
    //   })
    // ),
    checkedNodes: checkedNodesResult,
    intermediateNodes,
  };
}

/**
 *
 * 挂载子节点
 * @param {*} node
 * @param {*} childList
 * @param {*} treeProps
 */
export function loadNodeChild({
  node = null,
  childList = [],
  treeProps = {},
  checkedNodes,
  disabledNodeKeys,
} = {}) {
  if (!node) {
    throw new Error('param node is need.');
  }

  const { dynamicNodeLabel, dynamicNodeKey, staticNodeKey, dynamicNodeTypeLabel } = treeProps;
  const { _id: parentNodeId, _checked: parentNodeChecked, _disabled: parentNodeDisabled } = node;

  const filterChildren = node._children.filter(item => item.type !== 'user');
  node._children = filterChildren;

  childList.forEach(item => {
    node._children.push({
      ...item,
      _id: item[dynamicNodeKey],
      _name: item[dynamicNodeLabel],
      _children: [],
      _isLeaf: true,
      _checked:
        parentNodeChecked ||
        nodeInteractionState({
          type: 'checked',
          node: { ...item },
          nodeType: dynamicNodeTypeLabel,
          comparateProp: dynamicNodeKey,
          attachedComparateProp: staticNodeKey,
          checkedNodes,
        }),
      _intermediate: false,
      _type: dynamicNodeTypeLabel,
      _disabled:
        nodeInteractionState({
          type: 'disabled',
          node: { ...item },
          comparateProp: dynamicNodeKey,
          disabledNodeKeys,
        }) || parentNodeDisabled,
    });
  });

  // 更新节点状态
  node._isLoaded = true;
  node._isLeaf = !node._children.length;
  updateCheckedNodeKeys({ deptId: parentNodeId, dynamicNodeTypeLabel, checkedNodes });
}

/**
 * 更新本地缓存的_checkedNodeKeys，解决getChangedNodes错误添加的问题
 * @param {*} deptId 某个父节点的id
 */
export function updateCheckedNodeKeys({ deptId = '', dynamicNodeTypeLabel, checkedNodes } = {}) {
  if (!Array.isArray(checkedNodes) || !checkedNodes.length || typeof checkedNodes[0] === 'string') {
    return false;
  }

  checkedNodes.forEach(item => {
    const { type, deptId: _deptId = '' } = item;
    if (type === dynamicNodeTypeLabel && _deptId === deptId) {
      item._hasloaded = true;
    }
  });
}
